//##############################################################################
// Author : Aviral Mittal
// Author's Email : avimit@yahoo.com
// University of Edinburgh
//##############################################################################
//# This program reads a qcif file and converts it into a 
//# Verilog Readable format
//# The output looks like
//# ---------------
//# @0 xxxx_xxxx
//# @1 xxxx_xxxx
//# @2 xxxx_xxxx
//# @3 xxxx_xxxx
//# @4 xxxx_xxxx
//# .....
//# .....
//# ---------------
//##############################################################################

#define NULL (0)
#include <stdio.h>
#include <string.h>
#define roww 144
#define coll 176
#define cb_coll 72
#define cb_roww 88
#define cr_coll 72
#define cr_roww 88

//Function Prototypes Beigins --{
void HandleArgs(int argc, char * argv[], char * inpfile,int * fr_nos);
void PrintHelpMess();
void myfopen(FILE **fileptr,int mode, char * fname);
void PrintErr(int * fr_nos);
void GetFrames(
     FILE * fin,
     unsigned char lumaR[roww][coll],
     unsigned char lumaC[roww][coll],
     unsigned char cb[cb_roww][cb_coll],
     unsigned char cr[cr_roww][cr_coll],
     int row, 
     int col,
     int cb_row,
     int cb_col,
     int cr_row,
     int cr_col,
     int RefFrNo,
     int CurrFrNo
     );
void PrintArray(FILE * fr, 
                unsigned char chararray[roww][coll],
                int row, 
                int col);
void PrintHeader(FILE * fout,int row, int col);
void PrintInfo(FILE *fout,
               int rows,
               int cols,
               int rscope,
               int cscope,
               int brows,
               int bcols);
int main (int argc, char * argv[])
{
  //Variables Section Begins {
  int prn=1;
  unsigned char def_in_file[] = "default.qcif";
  char inpfile[250];
  char outfile[250];
  
  char rFrfile[11] = "rFrame.pgm";
  char cFrfile[11] = "cFrame.pgm";

  int stfr=0; //start frame number
  int finfr=0; //end frame number
  int fr_nos[2] = {0,0}; //For frame numbers from command line

  FILE * fin;//input file
  FILE * fout;//output file

  FILE * rFout;//output file
  FILE * cFout;//output file

  int row = roww;
  int col = coll;
  int cb_row = cb_roww;
  int cb_col = cb_coll;
  int cr_row = cr_roww;
  int cr_col = cr_coll;

  unsigned char lumaR[row][col];
  unsigned char lumaC[row][col];
  unsigned char cb[cb_row][cb_col];
  unsigned char cr[cr_row][cr_col];

  unsigned char lumaN[row][col];

  FILE * RecFileH;//output reconstructed frame
  char RecFrFile[13] = "RecFrame.pgm";

  int errFunc = 2; //1 for sad, 2 for mse
  //Variables Section Ends }

  strcpy(inpfile,def_in_file);
  strcpy(outfile,"out.dat");
  HandleArgs(argc,argv,inpfile,fr_nos);
  if(fr_nos[1] <= fr_nos[0]) PrintErr(fr_nos);
  //if(prn) printf("**************************************************\n");
  //if(prn) printf("INFO! Input File is = %s\n",argv[1]);
  //if(prn) printf("INFO! Reference Frame = %d\n",fr_nos[0]);
  //if(prn) printf("INFO! Current Frame = %d\n",fr_nos[1]);
  //if(prn) printf("**************************************************\n");
  if(prn) printf("**************************************************\n");
  if(prn) printf(" %s ",argv[1]);
  if(prn) printf(" Ref Frame = %d ",fr_nos[0]);
  if(prn) printf(" CurFrame = %d ",fr_nos[1]);
  if(prn) printf(" delta = %d \n",fr_nos[1]-fr_nos[0]);
  //if(prn) printf("**************************************************\n");
  myfopen(&fin,0,inpfile); //0 means open as rb
  myfopen(&fout,2,outfile); //2 means open as w
  GetFrames(fin,lumaR,lumaC,cb,cr,row,col,cb_row,cb_col,cr_row,cr_col,fr_nos[0],fr_nos[1]);
  myfopen(&cFout,2,cFrfile); //2 means open as w
  PrintArray(cFout,lumaC,row,col);
  myfopen(&rFout,2,rFrfile); //2 means open as w
  PrintArray(rFout,lumaR,row,col);
}// int main (int argc, char * argv[])


void HandleArgs(int argc, char * argv[], char * inpfile, int * fr_nos)
{
//###################################
//# This function Hs input args
//###################################
int stfr = 0;//local copy of it into this function.
int finfr = 0;//local copy of it into this function.
switch(argc)
  {
    case 1 :
      { 
        PrintHelpMess();
        exit(-1);
      }// case 1 :
    case 2 :
      { 
        //inpfile=argv[1];
        PrintHelpMess();
        exit(-1);
        break;
      }// case 2 :
    case 3 :
      {
        strcpy(inpfile,argv[1]);
        //printf("INFO! #1. Input file is %s\n",inpfile);
        //grab start frame and end frame numbers in variables.
        //The input argv is a string. so a string of type
        //say "23-45" which contains the start and end frame to be printed
        //must be processed to get integers stfr, start frame number
        //finfr, finish frame number.
        while(*argv[2] != '-')
        {
          stfr = *argv[2]-48+stfr; //Convert argv[2] to an integer stfr
          argv[2]++;
          stfr=stfr*10;
        }
        stfr=stfr/10; //Re-adjust the result
        argv[2]++;
        while(*argv[2])
        {
          finfr = *argv[2]-48+finfr;
          argv[2]++;
          finfr=finfr*10;
        }
        finfr=finfr/10; //Re-adjust the result
        fr_nos[0] = stfr;
        fr_nos[1] = finfr;
        break;
      }// case 3 :
    default: 
      {
        printf("ERROR! Too many args at command line. Exiting....\n");
        exit(-1);
      }
  }// switch(argc)
}// void HandleArgs(int argc, char * argv[], char * inpfile, int *fr_nos)
void PrintHelpMess()
{
        printf("###################################################\n");
        printf("#This Program extracts 2 named frames from qcif file,\n");
        printf("#Calculates motion estimation vectors,reconstructs \n");
        printf("#Frame using motion Vectors \n");
        printf("#And Finally calculates PSNR of the Reconstructed Frame\n");
        printf("###################################################\n");
        printf("Usage: unix> a.out <in file> <n1-n2>\n");
        printf("Where n1 is the reference frame number\n");
        printf("Where n2 is the frame to be encoded \n");
        printf("Example: unix> a.out bridge.qcif 45-47\n");
        printf("WARN!: Input args Format not checked. Pls conform to above shown usage\n");
        printf("###################################################\n");
}// void PrintHelpMess()
void myfopen(FILE **fileptr,int mode, char * fname)
{
//###################################
//# This function opens a file in
//# Binary Read Mode if mode = 0
//# Ascii Read Mode if mode = 1
//# Write Mode if mode = 2
//###################################
  //unsigned char luma[144][176];
  if(mode == 0)
  {
    if(!(*fileptr = fopen(fname, "rb")))
    {
      printf("1. Cannot open file :%s Exiting . . . .\n",fname);
      exit(-1);
    }// if(!(fileptr = fopen(fname, "r")))
  }// if(mode == 0)
  else if(mode == 1)
  {
    if(!(*fileptr = fopen(fname, "r")))
    {
      printf("2. Cannot open file :%s Exiting . . . .\n",fname);
      exit(-1);
    }// if(!(fileptr = fopen(fname, "r")))
  }// else if(mode == 1)
  else if(mode == 2)
  {
    if(!(*fileptr = fopen(fname, "w")))
    {
      printf("3. Cannot open file :%s Exiting . . . .\n",fname);
    exit(-1);
    }// if(!(fileptr = fopen(fname, "r")))
  }// else if(mode == 2)
  else
  {
    printf("############################################################\n");
    printf("ERROR! from function myfopen: Pls Dont Play with me\n");
    printf("Allowed Mode values are 0,1,2 where as mode passed is %d\n",mode);
    printf("Exiting....\n");
    printf("############################################################\n");
    exit(-1);
  }
}// void myfopen(FILE **fileptr,int mode, char * fname);

void GetFrames(
     FILE * fin,
     unsigned char lumaR[roww][coll],
     unsigned char lumaC[roww][coll],
     unsigned char cb[cb_roww][cb_coll],
     unsigned char cr[cr_roww][cr_coll],
     int row, 
     int col, 
     int cb_row,
     int cb_col,
     int cr_row,
     int cr_col,
     int RefFrNo,
     int CurrFrNo
     )
{
  int frno=0;
  int cc=0;
  int rr=0;
  for(frno=0;frno<RefFrNo+1;frno++)
  {
    fread(lumaR,1,row*col,fin); //read whole luma frame
    fread(cb,1,cb_row*cb_col,fin); //read whole luma frame
    fread(cr,1,cr_row*cb_col,fin); //read whole luma frame
  }// for(frno=0;frno<CurrFrno;frno++)
  for(;frno<CurrFrNo+1;frno++)
  {
    fread(lumaC,1,row*col,fin); //read whole luma frame
    fread(cb,1,cb_row*cb_col,fin); //read whole luma frame
    fread(cr,1,cr_row*cb_col,fin); //read whole luma frame
  }// for(frno=0;frno<CurrFrno;frno++)
}// void GetFrames()

void PrintArray(FILE * fr,
                unsigned char chararray[roww][coll],
                int row, 
                int col)
{
int rr=0;
int cc=0;
int ii=0;
//PrintHeader(fr,row,col);
for(rr=0;rr<row;rr++)
  for(cc=0;cc<col;cc++)
  {
    int jj;
    fprintf(fr,"@%X 0000_0000_0000_0000_0000_0000_",ii);
    for(jj=0;jj<8;jj++) 
    {
      if(jj==4) fprintf(fr,"_");
      fprintf(fr,"%d",( (chararray[rr][cc]>>(7-jj)) & 1) );
    }
    ii++;
    fprintf(fr,"\n");
  }// for(cc=0;cc<col;cc++)
}// void PrintArray()

void PrintHeader(FILE * fout,int row,int col)
{
  fprintf(fout,"P2\n");
  fprintf(fout,"%d %d\n",col,row);
  fprintf(fout,"255\n");
}// void PrintHeader(char * fname)

void PrintErr(int * fr_nos)
{
  printf("#################################################\n");
  printf("ERROR! Dont play around!!! Curr Frame Number %d must be greater than Ref Frame Number %d\n",fr_nos[1],fr_nos[0]);
  printf("Exiting....\n");
  printf("#################################################\n");
  exit(-1);
}// void PrintErr(int * fr_nos)

