////////////////////////////////////////////////////////////////////////////
//# Author Aviral Mittal
//# ISLI 2005/2006
//# File Name mes.v
////////////////////////////////////////////////////////////////////////////
////
`timescale 1ns/1ps
//`include "./functions.v"

module mes(mv_x,
           mv_y,
           min_sad,
           partition_mode,
           mfinish,
           ADRRef,
           CENRef,
           WENRef,
           OENRef,
           ADRCur,
           CENCur,
           WENCur,
           OENCur,
           ADRRec ,
           CENRec ,
           WENRec,
           OENRec,
           DRec,
           start,
           clk,
           reset,
           Dref,
           Dcur
           );

`include "../rtl/functions.v"
`ifdef CIF // SIMONLY
`include "../rtl/cif_parameters.v" // SIMONLY
`else // SIMONLY
`include "../rtl/qcif_parameters.v"
`endif // SIMONLY


  input start;
  input clk;
  input reset;
  input [data_width-1:0] Dref;
  input [data_width-1:0] Dcur;

  output [7:0]  mv_x;
  output [7:0]  mv_y;
  output [15:0] min_sad;
  output [9:0]  partition_mode;
  output        mfinish;

  output [16:0] ADRRef;
  output CENRef;
  output WENRef;
  output OENRef;

  output [16:0] ADRCur;
  output CENCur;
  output WENCur;
  output OENCur;

  output [16:0] ADRRec;
  output CENRec;
  output WENRec;
  output OENRec;

  output [data_width-1:0] DRec;


  reg [rrwidth-1:0] rr;
  reg [rrwidth-1:0] n_rr;
  reg [ccwidth-1:0] cc;
  reg [ccwidth-1:0] n_cc;

  reg [l2broww-1:0] brow;
  reg [l2broww-1:0] n_brow;
  reg [l2bcoll-1:0] bcol;
  reg [l2bcoll-1:0] n_bcol;

  reg [l2rscopee-1:0] rscope;
  reg [l2rscopee-1:0] n_rscope;
  reg [l2cscopee-1:0] cscope;
  reg [l2cscopee-1:0] n_cscope;

  reg [l2broww-1:0] iir;
  reg [l2broww-1:0] n_iir;
  reg [l2bcoll-1:0] iic;
  reg [l2bcoll-1:0] n_iic;

  reg [l2broww-1:0] iir_p0;
  reg [l2bcoll-1:0] iic_p0;
  reg [l2broww-1:0] iir_p1;
  reg [l2bcoll-1:0] iic_p1;

  reg [l2roww-1:0] refrr;
  reg [l2roww-1:0] n_refrr;
  reg [l2coll-1:0] refcc;
  reg [l2coll-1:0] n_refcc;
  reg [l2coll-1:0] frefrr;
  reg [l2coll-1:0] frefcc;

  reg [l2coll-1:0] frefrr_p0;
  reg [l2coll-1:0] frefcc_p0;

  reg [1:0] iter;
  reg [1:0] n_iter;

  reg [3:0] ii;
  reg [3:0] n_ii;

  reg [17:0] SAD;
  reg [17:0] MinSAD;
  reg [l2rscopee-1:0] xvec;
  reg [l2cscopee-1:0] yvec;
  reg [17:0] n_SAD;
  reg [17:0] n_MinSAD;
  reg [l2rscopee-1:0] n_xvec;
  reg [l2cscopee-1:0] n_yvec;

  wire [16:0] refAddr;
  wire [16:0] curAddr;
 
  wire [l2roww-1:0] crow;
  wire [l2coll-1:0] ccol;

  reg [rrwidth-1:0] n_crow;
  reg [rrwidth-1:0] n_ccol;
 
  reg ld_crow;
  reg ld_ccol;

  wire row_err;
  wire row_err_under;
  wire row_err_over;
  wire col_err;
  wire col_err_under;
  wire col_err_over;

  wire addr_err;
  wire addr_err_mod;

  reg addr_err_del;
  wire addr_err_pulse_n;
  wire addr_err_pulse;

  reg max_iic_del; //For generating a puslse of iic==brow, to make sure
                   //this sinal is '1' for only one clock cycle
                   //This will 'stall' the design properly
                   //without if, if start goes to '0', when iic==3,
                   //The design will keep advancing
  wire max_iic_pulse; //For generating a puslse of iic==brow, to make sure
                   //this sinal is '1' for only one clock cycle
                   //This will 'stall' the design properly
                   //without if, if start goes to '0', when iic==3,
                   //The design will keep advancing
  reg start_del;
  reg start_del_del;
  wire start_pulse;
  reg start_pulse_del; //For making sure that at start up
                       //SAD reg do not take up a X value.
                       //Since it would take 1 clock cycle 
                       //AFter issuing start, to get a valid data
                       //from memory, during this clock cycle
                       //Data from memoryies will be xxxxxxxxx

  reg mindex;
  reg n_mindex;
  wire OneMatchDone;
  reg OneMatchDone_del;
  reg OneMatchDone_del_del;
  wire OneSearchComplete;
  reg OneSearchComplete_del;
  reg OneSearchComplete_del_del;
  wire OneIterDone;
  wire OneFrameDone;
  reg OneFrameDone_del;
  reg OneFrameDone_del_del;

  wire incr_iir;
  wire incr_iic;
  wire incr_rr;
  wire incr_cc;
  wire incr_refrr;
  wire incr_refcc;
  wire incr_iter;
  wire incr_ii;
  wire incr_mindex;
  wire ld_refrr;
  wire ld_refcc;

  wire max_iir;
  wire max_iic;
  wire max_rr;
  wire max_cc;
  wire max_refrr;
  wire max_refcc;
  wire max_iter;
  wire max_ii;

  wire [data_width:0] delta; // DRef-Dcur
  wire [data_width:0] abs_delta; // abs(DRef-Dcur)
  wire [data_width:0] col_m_bcol;
  wire [data_width:0] row_m_brow;

  reg [rrwidth-1:0] rr_p0;
  reg [ccwidth-1:0] cc_p0;
  reg [rrwidth-1:0] rr_p1;
  reg [ccwidth-1:0] cc_p1;
  
  reg rec_mem_wr;
  wire rec_mem_wr_int;

  //reg fifo_rd;
  //wire fifo_rd_int;
  

  wire fifo_wr; //Not assigned yet avi ##
  wire fifo_rd;

  reg wr_block; 
  reg n_wr_block;
  reg wr_block_del; 

//   function [16:0] mlog;
//   input [4:0] mvec;
//   begin
//     case(mvec)
//       4 : mlog = 2;
//       8 : mlog = 3;
//       16 : mlog = 4;
//       32 : mlog = 5;
//       default : mlog=0;
//     endcase
//   end
//   endfunction
  

//Reset all registers  
  always @ (posedge clk)
  begin
    if(reset == 0)
    begin
      rr <= 0;
      cc <= 0;
      iir <= 0;
      iic <= 0;
      brow <= broww;
      bcol <= bcoll;
      refrr <= 0;
      refcc <= 0;
      ii <= 0;
      iter <= 0;
      //crow <= 0;
      //ccol <= 0;
      max_iic_del <= 0;
      start_del <= 0;
      start_del_del <= 0;
      SAD <= 0;
      MinSAD <= {(17 ){1'b1}} ;
      xvec <= 0;
      yvec <= 0;
      rec_mem_wr <= 0;
      wr_block <= 0;
      wr_block_del <= 0;
    end
    else
    begin
      rr <= n_rr;
      cc <= n_cc;
      iir <= n_iir;
      iic <= n_iic;
      refrr <= n_refrr;
      refcc <= n_refcc;
      ii <= n_ii;
      iter <= n_iter;
      //crow <= n_crow;
      //ccol <= n_ccol;
      max_iic_del <= max_iic;
      start_del <= start;
      start_del_del <= start_del;
      start_pulse_del <= start_pulse;
      SAD <= n_SAD;
      MinSAD <= n_MinSAD ;
      addr_err_del <= addr_err;
      OneMatchDone_del <= OneMatchDone;
      OneMatchDone_del_del <= OneMatchDone_del;
      OneSearchComplete_del <= OneSearchComplete;
      OneSearchComplete_del_del <= OneSearchComplete_del;
      xvec <= n_xvec;
      yvec <= n_yvec;
      iir_p0 <= iir;
      iic_p0 <= iic;
      iir_p1 <= iir_p0;
      iic_p1 <= iic_p0;
      rr_p0 <= rr;
      cc_p0 <= cc;
      rr_p1 <= rr_p0;
      cc_p1 <= cc_p0;
      frefrr_p0 <= frefrr;
      frefcc_p0 <= frefcc;
      OneFrameDone_del <= OneFrameDone;
      OneFrameDone_del_del <= OneFrameDone_del;
      rec_mem_wr <= rec_mem_wr_int;
      wr_block <= n_wr_block;
      wr_block_del <= wr_block;
    end 
  end
//Reset all registers Ends

  assign start_pulse = start & ~start_del;

// Combo Logic for rr

  always @ (iic or incr_iic)
  begin
    n_iic <= iic;
    if(incr_iic)
      if(iic == bcoll -1)
        n_iic <= 0;
      else
        n_iic <= iic + 1;
  end

  always @ (iir or incr_iir)
  begin
    n_iir <= iir;
      if(incr_iir)
        if(iir == broww -1)
          n_iir <= 0;
        else
          n_iir <= iir + 1;
  end

  always @ (rr or incr_rr)
  begin
    n_rr <= rr;
    if(incr_rr)
      if(rr == roww/broww -1)
        n_rr <= 0;
      else
        n_rr <= rr + 1;
  end


  always @ (cc or incr_cc)
  begin
    n_cc <= cc;
    if(incr_cc)
      if(cc == coll/bcoll -1)
        n_cc <= 0;
      else
        n_cc <= cc + 1;
  end

  always @ (ii or incr_ii)
  begin
    n_ii <= ii;
    if(incr_ii)
      if(ii != 8)
        n_ii <= ii + 1;
      else
        n_ii <= 0;
  end


// if(ii==1) frefrr = refrr+brow/2;
// if(ii==1) frefcc = refcc+bcol/2;
// if(ii==3) frefrr = refrr+brow/2;
// if(ii==3) frefcc = refcc-bcol/2;
// if(ii==7) frefrr = refrr-brow/2;
// if(ii==7) frefcc = refcc+bcol/2;
// if(ii==9) frefrr = refrr-brow/2;
// if(ii==9) frefcc = refcc-bcol/2;

  //refrr,refcc
  assign crow = n_rr<<(mlog(brow));
  assign ccol = n_cc<<(mlog(bcol));
  always @ (refrr or incr_refrr or n_iter or n_ii or iir or crow or brow or ld_refrr)
  begin
    n_refrr <= refrr;
    if(ld_refrr == 1 && n_iter == 0 && n_ii==0)
      n_refrr <= crow-brow;
    else if(ld_refrr == 1 && n_iter != 0 && n_ii==0)
      n_refrr <= crow-1;
    else if(incr_refrr == 1 && n_iter == 0)
      if(refrr == crow+brow)
        n_refrr <= crow-brow;
      else
        n_refrr <= refrr+brow;
    else if(incr_refrr && n_iter !=0) 
      if(refrr == crow+1)
        n_refrr <= crow-1;
      else
        n_refrr <= refrr+1;
  end

  always @ (refrr or ii or iter or brow)
  begin
    frefrr <= refrr;
    if((ii==0 || ii==2) && iter==0)
      frefrr <= refrr+(brow>>1);
    else if((ii==6 || ii==8) && iter==0)
      frefrr <= refrr-(brow>>1);
  end

  always @ (refcc or incr_refcc or n_iter or n_ii or iir or ccol or bcol or ld_refcc)
  begin
    n_refcc <= refcc;
    if(ld_refcc == 1 && n_iter == 0 && n_ii==0)
      n_refcc <= ccol-bcol;
    else if(ld_refcc == 1 && n_iter != 0 && n_ii==0)
      n_refcc <= ccol-1;
    else if(incr_refcc == 1 && n_iter == 0)
      if(refcc == ccol+bcol)
        n_refcc <= ccol-bcol;
      else
        n_refcc <= refcc+bcol;
    else if(incr_refcc && n_iter != 0) 
      if(refcc == ccol+1)
        n_refcc <= ccol-1;
    else
        n_refcc <= refcc+1;
  end

  always @ (refcc or ii or iter or bcol)
  begin
    frefcc <= refcc;
    if((ii==0 || ii==6) && (iter==0))
      frefcc <= refcc+(bcol>>1);
    else if((ii==2 || ii==8) && (iter==0))
      frefcc <= refcc-(bcol>>1);
  end


//  always @(ccol or ld_ccol)
//  begin
//    n_ccol <= ccol;
//    if(ld_ccol)
//      n_ccol <= cc*bcol;
//  end

//  always @(crow or ld_crow)
//  begin
//    n_crow <= crow;
//    if(ld_crow)
//      n_crow <= rr*brow;
//  end

  always @(iter or incr_iter)
  begin
    n_iter <= iter;
    if(incr_iter)
      if(iter == 2)
        n_iter <= 0;
      else
        n_iter <= iter+1;
  end

  always @ (OneMatchDone or mindex)
  begin
    n_mindex <= mindex;
    if(OneMatchDone && mindex == 26)
      n_mindex <= 0;
    else if(OneMatchDone)
      n_mindex <= n_mindex+1;
  end


  assign max_iic = (iic==bcol-1) ? 1 : 0;
  assign max_iir = (iir==brow-1) ? 1 : 0;
  assign max_iter = (iter==3-1) ? 1 : 0;
  assign max_ii = (ii==9-1) ? 1 : 0;

  assign max_cc = (cc==(coll>>mlog(bcol))-1) ? 1 : 0;
  assign max_rr = (rr==(roww>>mlog(brow))-1) ? 1 : 0;

  assign max_iic_pulse = max_iic & ~max_iic_del;
  assign incr_iic=start_del;
  assign incr_iir = max_iic_pulse;

  assign incr_ii   = OneMatchDone;
  assign incr_iter = OneIterDone;

//  always @(max_iter or max_ii or max_refcc or max_refrr or max_refrr or max_iir or max_iic)
//  beign
//    incr_iir = 0;
//    incr_refrr = 0;
//    incr_refcc = 0;
//    incr_iter = 0;
//    incr_ii = 0;
//    incr_rr = 0;
//    incr_cc = 0;
//    if(incr_iic)
//    begin
//    end
//        
//  end

  assign OneMatchDone = max_iic_pulse & max_iir;
  assign OneIterDone  = OneMatchDone & max_ii;
  assign OneSearchComplete = OneIterDone & max_iter;
  assign incr_cc = OneSearchComplete;
  assign incr_rr = incr_cc & max_cc;
  assign OneFrameDone = incr_rr & max_rr;


  assign ld_refcc = start_pulse | OneMatchDone; 
  assign ld_refrr = start_pulse | OneMatchDone; 

  assign incr_refcc = OneMatchDone;
  assign incr_refrr = OneMatchDone & (ii==2 || ii==5 || ii==8);

  assign col_m_bcol = coll-bcol;
  assign row_m_brow = roww-brow;

  assign row_err_under = (refrr[l2roww-1]) ? 1 : 0;
  assign row_err_over = (refrr > row_m_brow) ? 1 : 0;
  assign col_err_under = (refcc[l2coll-1]) ? 1 : 0;
  assign col_err_over = (refcc > col_m_bcol) ? 1 : 0;
  assign col_err = col_err_under || col_err_over;
  assign row_err = row_err_under || row_err_over;


  //ACU
  //addr = ((rr*brow)+iir)*col+(cc*bcol)+iic); //from Reference C Model
  assign curAddr = ((((rr<<mlog(brow))|iir)+ (((rr<<mlog(brow))|iir)<<1)+(((rr<<mlog(brow))|iir)<<3))<<cif_qcif) + ((cc<<mlog(bcol))|iic);

  assign refAddr = (((frefrr+iir)+ ((frefrr+iir)<<1)+((frefrr+iir)<<3))<<cif_qcif) + (frefcc+iic);

  assign addr_err = row_err | col_err;

  assign ADRRef = (addr_err) ? 0 : refAddr;
  assign ADRCur = (addr_err) ? 0 : curAddr;
 
  assign CENRef = ~start | start_pulse;
  assign OENRef = ~start | start_pulse;
  assign WENRef = 1'b1;

  assign CENCur = ~start | start_pulse;
  assign OENCur = ~start | start_pulse;
  assign WENCur = 1'b1;


  assign delta = {1'b0,Dref}-{1'b0,Dcur}; //make unsigned
  assign abs_delta = delta[data_width] ? (~delta+1): delta;
  assign addr_err_pulse_n = addr_err_del & ~addr_err;
  assign addr_err_pulse = addr_err & ~addr_err_del;
  assign addr_err_mod = addr_err & ~addr_err_pulse;

  always @ (SAD or 
            MinSAD or 
            addr_err or 
            start or 
            start_pulse or 
            start_pulse_del or 
            delta or 
            addr_err or 
            addr_err_pulse or 
            addr_err_pulse_n or 
            addr_err_mod or 
            OneMatchDone_del or 
            OneMatchDone_del_del or 
            OneSearchComplete_del or 
            OneSearchComplete_del_del or 
            frefrr_p0 or 
            frefcc_p0 or 
            iir_p0 or 
            iic_p0 or 
            rr_p0 or 
            cc_p0 or 
            abs_delta or
            brow or
            wr_block or
            xvec or
            yvec or
            bcol)
  begin
      n_SAD <= SAD; 
      n_MinSAD <= MinSAD;
      n_xvec <= xvec;
      n_yvec <= yvec;
      n_wr_block <= wr_block;
    if(OneMatchDone_del)
        n_wr_block <= 0;
    if(start == 1 && start_pulse == 0 && start_pulse_del == 0 && addr_err == 0 && addr_err_pulse_n == 0 && OneMatchDone_del_del==1)
    begin
      n_SAD <= abs_delta;
    end
    else if(start == 1 && start_pulse == 0 && start_pulse_del == 0 && addr_err_mod == 0 && addr_err_pulse_n == 0)
    begin
      n_SAD <= SAD + abs_delta;
    end

    if(start == 1 && start_pulse == 0 && start_pulse_del == 0 && addr_err_mod == 0 && addr_err_pulse_n == 0 && OneMatchDone_del==1)
    begin
      if((SAD + abs_delta) < MinSAD)
      begin
        n_MinSAD <= SAD + abs_delta;
        n_xvec <= frefrr_p0-(rr_p0<<mlog(brow));
        n_yvec <= frefcc_p0-(cc_p0<<mlog(bcol));
        n_wr_block <= 1;
      end
    end
    if(OneSearchComplete_del_del)
      n_MinSAD <= {(17 ){1'b1}} ;
  end


  assign mfinish = OneFrameDone_del_del;

  always @ (posedge clk)
  begin
    //if(addr_err == 0 && it == 0)
    //$display("ADRRef= %d  ADRCur= %d it=%d err=%d rr=%d cc=%d time=%d", ADRRef,ADRCur,iter,addr_err,rr,cc,$time);
    if(OneSearchComplete_del_del)
      $display("MinSAD= %d  for (%d %d) X %d Y %d time=%d",MinSAD,rr_p0,cc_p0,xvec,yvec,$time);
    //jj<=jj+1;
    //$writememb("./mem_file1.v",ADRRef);
    //$writememb("./mem_file1.v",ADRCur);
  end

//////////////////////////////
//# Fifo Instantiation
//////////////////////////////


assign rec_mem_wr_int = (iter==2 && ii == 4) ? 1 : 0;
//assign fifo_rd = rec_mem_wr_int;
assign fifo_rd = wr_block;

assign WENRec = ~wr_block_del;
assign CENRec = ~wr_block_del;
assign OENRec = ~wr_block_del;
//assign ADRRec = CurAdr_del_del;
assign ADRRec = ((((rr_p1<<mlog(brow))|iir_p1)+ 
                 (((rr_p1<<mlog(brow))|iir_p1)<<1)+
                 (((rr_p1<<mlog(brow))|iir_p1)<<3))<<cif_qcif) + 
                 ((cc_p1<<mlog(bcol))|iic_p1);


assign fifo_wr = (addr_err_del==0 && start_del_del==1) ? 1 : 0 ; //if there is a vlaid data being
                                //fetched, it must be recorded

assign mv_x = xvec;
assign mv_y = yvec;
assign min_sad = MinSAD;

 fifo_lp fifo_lp_i0 (
.datout(DRec[data_width-1:0]),
.full(),
.empty(),
.clk(clk),
.wr(fifo_wr),
.rd(fifo_rd),
.rst_n(reset),
.datin(Dref[data_width-1:0])
);


//  always @ (posedge clk)
//    if(OneSearchComplete)
//    begin
//      rr_p0 <= rr;
//      cc_p0 <= cc;
//    end
endmodule
