#!/bin/csh -f
set title = `basename $0`; 
########################### 
# FOR ARGUMENTS 
########################### 
  if ($#argv < 3) then 
     set error = "Invalid number of arguments;At least 3 expected" 
     echo "##########################################################" 
     echo "ERROR ($title) :"$error; 
     echo "USAGE: unix> prj.csh <qcif_file> <st_fr> <end_fr> [gui/net] "
     echo "where stfr is the ref Frame No. and end_fr is Cur Frame No."
     echo "where option gui/net is optional, and will stat gui or net sims"
     echo "Example: unix> prj.csh ../video/carphone.qcif 34 36"
     echo "##########################################################"
     exit 
  endif 

set C_SRC = "../c_src"
set BIN = "../bin"
set SIM = "../sim"
set VIDEO_SRC = "../video"
set qcif_file = $argv[1]
set RTL = ../rtl
set TB = ../tb
set RESULTS = ../results
set SYN = ../synopsys
set SCRIPTS = ../scripts
set NETLIST = ../netlist
set POWER = ../power

rm -rf $SIM/mes.log
rm -rf $SIM/qcif2hdlinp.log

# start frame, servers as reference frame
set st_fr  = $argv[2]
# end frame, servers as current frame to be encoded
set end_fr = $argv[3] 
if(`uname` == "Linux") then
  source $SCRIPTS/source_file
else
  source $SCRIPTS/source_file_solaris
endif
##############################################################################
# Check and compile c file to convert qcif/cif file to rtl readable format
##############################################################################
  if(-r $C_SRC/qcif2hdlinp.c) then
    gcc $C_SRC/qcif2hdlinp.c -lm -o $BIN/qcif2hdlinp
  else
    echo "ERROR! Source File qcif2hdlinp.c Not found in #$C_SRC# ....Exiting" 
    exit(0)
  endif
##############################################################################
  
##############################################################################
# Check and compile c file to which performs ME on a givne qcif/cif file
##############################################################################
  if(-r $C_SRC/mes.c) then
    gcc $C_SRC/mes.c -lm -o $BIN/mes
  else
    echo "ERROR! Source File mes.c Not found in #$C_SRC# ....Exiting" 
    exit(0)
  endif
##############################################################################

##############################################################################
# Run C program mes to perform motion estimation on a qcif/cif file
# To generate reference motion vectors to be used for RTL verification
##############################################################################
  if(-r $VIDEO_SRC/$qcif_file) then
  else
    echo "ERROR! # $qcif_file # Not found in #$VIDEO_SRC# ....Exiting" 
    exit(0)
  endif

  if(-r $BIN/mes) then
    $BIN/mes $VIDEO_SRC/$qcif_file $st_fr-$end_fr
    echo "Running C Program for Motion Estimaion ......."
    mv ./out.dat $SIM
  else
    echo "ERROR! #mes# Executable NOT found #$BIN# ....Exiting" 
    exit(0)
  endif
##############################################################################

##############################################################################
# Run C program to generate input data for RTL Simulation
# This is effectivly to produce 2 frames froma qcif file
##############################################################################
  if(-r $BIN/qcif2hdlinp) then
    $BIN/qcif2hdlinp $qcif_file $st_fr-$end_fr > $SIM/qcif2hdlinp.log
    echo "Generating Simulation Data ......."
    mv cFrame.pgm $SIM
    mv rFrame.pgm $SIM
  else
    echo "ERROR! #qcif2hdlinp# Executable NOT found #$BIN# ....Exiting" 
    exit(0)
  endif
##############################################################################

##############################################################################
# Run HDL SiM
##############################################################################
  if(-r $RTL/mes.v && -r $TB/mes_tb.v) then
    echo "Runnign Simulation ......."
    rm -rf $SIM/work
    pushd $SIM
    vlib work
    vlog $RTL/mes.v $TB/mes_tb.v $TB/RamRef.v $TB/RamCur.v $RTL/fifo_lp.v $TB/RamRec.v
    if($#argv > 3) then
      if($argv[4] == "gui") then
        vsim -do "add wave sim:/mes_tb/mes/*" mes_tb
      else if($argv[4] == "net") then
        echo "NETLIST SIMULATION"
        source $SCRIPTS/run_net.scr
      else
        vsim -c -do "run 7 ms; force -freeze sim:/mes_tb/R64X32M4_Rec/dump 1 0; run 100ns; quit" mes_tb
      endif
    else
      vsim -c -do "run 7 ms; force -freeze sim:/mes_tb/R64X32M4_Rec/dump 1 0; run 100ns; quit" mes_tb
    endif
    popd
  else
    echo "ERROR! #Missing mse.v or mse_tb.v# in #$RTL $TB res#....Exiting" 
    exit(0)
  endif
##############################################################################


rm -rf $SIM/hdl_sad $SIM/c_sad sad_diff
pushd $SIM
grep MinSAD transcript | awk '{print $3}' > hdl_sad
grep minMSE out.dat | awk '{print $4}' > c_sad
diff hdl_sad c_sad > $RESULTS/sad_diff
popd

echo "########################################################################"
echo "# RTL vs C results are in $RESULTS/sad_diff                           #"
echo "########################################################################"

rm -rf $RESULTS/hdl.pgm $RESULTS/temp $SIM/diff
sed -n "4,$ p" $SIM/RecFrame.pgm | sed "s/z//g" > $RESULTS/temp
gcc $C_SRC/hex2int.c -o $BIN/hex2int
$BIN/hex2int > $RESULTS/hdl.pgm
gcc $C_SRC/cal_sad_psnr.c -o $BIN/sad_psnr -lm
$BIN/sad_psnr $VIDEO_SRC/$qcif_file $st_fr-$end_fr
mv *.pgm $RESULTS/
mv out.dat $RESULTS/
diff $RESULTS/hdl.pgm $RESULTS/RecFrame.pgm > $RESULTS/recFr_diff

echo "########################################################################"
echo "# RTL vs C results are in $RESULTS/recFr_diff                          #"
echo "########################################################################"

cp $SYN/netlist.v $NETLIST
cp $SYN/power.rep $POWER

set uname = `uname`

if($uname == "Linux") then
  xview $RESULTS/hdl.pgm &
  xview $RESULTS/RecFrame.pgm &
  xview $RESULTS/cFrame.pgm &
else
  xv $RESULTS/hdl.pgm &
  xv $RESULTS/RecFrame.pgm &
  xv $RESULTS/cFrame.pgm &
endif


