//##############################################################################
// Author : Aviral Mittal
// Author's Email : avimit@yahoo.com
// University of Edinburgh
// File name : cal_psnr.c
//##############################################################################
//# Release Date : 25Aug2006
//# Version      : 1.0
//# Notes        : This program is capable of calculating motion vectors from a
//#              : QCIF file. Then reconstructing a Frame using those motion
//#              : vectors and refreence frame.And Then finding out what is the
//#                Peak Signal to Noise Ratio of the reconstructed frame
//#                with respect to the reference frame
//#
//# Limitations  : It cannot check if the reference frame and current frame are
//#                present in the qcif file given as input. It is the 
//#                responsibility of the user to behave :).
//#                The Program doesnt make any checks if the input file format
//#                is corret.
//# Usage        : Compile it using gcc and try to execute a.out without any 
//#                args. It will print out helpful message on how to use this
//#                program
//#                unix> a.out <qcif file> <n1-n2> 
//#                unix> a.out container.qcif 45-49
//#                where n1 is the reference frame and n2 is the current frame
//#                for which it is intended to obtain motion vectors
//# Outputs      : This program produces 4 files as outputs
//#                1). rFrame.pgm  : contains the reference frame, can be viewed
//#                                  by xv
//#                2). cFrame.pgm  : contains the current frame, can be viewed
//#                                  by xv
//#                3). out.dat     : contains the Reference Block Number, 
//#                                  corresponding SAD, corresponding motion 
//#                                  vectors
//#                4). RecFrame.pgm: contains the Reference Block Number, 
//#                
//#                This program also prints PSNR at stdout 
//##############################################################################

#define NULL (0)
#include <stdio.h>
#include <string.h>
#define roww 144
#define coll 176
#define cb_coll 72
#define cb_roww 88
#define cr_coll 72
#define cr_roww 88
#define itrations 4

//legal allowed values for broww in H264 is 4,8,16
//Following must be such that mod(roww/broww)=0
#define broww 4
//Following must be such that mod(coll/b4oll)=0
#define bcoll 4
//Following must be such that mod(broww/rscopee=0)
#define rscopee 8
//Following must be such that mod(bcoll/cscopee=0)
#define cscopee 8



//Function Prototypes Beigins --{
void HandleArgs(int argc, char * argv[], char * inpfile,int * fr_nos);
void PrintHelpMess();
void myfopen(FILE **fileptr,int mode, char * fname);
void PrintErr(int * fr_nos);
void GetFrames(
     FILE * fin,
     unsigned char lumaR[roww][coll],
     unsigned char lumaC[roww][coll],
     unsigned char cb[cb_roww][cb_coll],
     unsigned char cr[cr_roww][cr_coll],
     int row, 
     int col,
     int cb_row,
     int cb_col,
     int cr_row,
     int cr_col,
     int RefFrNo,
     int CurrFrNo
     );
void PrintArray(FILE * fr, 
                unsigned char chararray[roww][coll],
                int row, 
                int col);
void PrintHeader(FILE * fout,int row, int col);
void PrintInfo(FILE *fout,
               int rows,
               int cols,
               int rscope,
               int cscope,
               int brows,
               int bcols);
void GetVecs(unsigned char lumaR[roww][coll],
             unsigned char lumaC[roww][coll],
             FILE * fout,
             int rows,
             int cols,
             int rscopes,
             int cscopes,
             int brows,
             int bcols,
             char xvec[roww/broww][coll/bcoll],
             char yvec[roww/broww][coll/bcoll],
             int errFunc);

void ReConstruct(char xvec[roww/broww][coll/bcoll],
                 char yvec[roww/broww][coll/bcoll],
                 unsigned char lumaR[roww][coll],
                 unsigned char lumaN[roww][coll], //Reconstructed Frame
                int rows, 
                int cols, 
                int brows, 
                int bcols
                );
void CalPsnr(
            unsigned char lumaC[roww][coll],
            unsigned char lumaN[roww][coll],
            int rows,
            int cols,
            int bcols,
            int brows,
            int errFunc
            );
//Function Prototypes Ends --}

int main (int argc, char * argv[])
{
  //Variables Section Begins {
  int prn=1;
  unsigned char def_in_file[] = "default.qcif";
  char inpfile[250];
  char outfile[250];
  
  char rFrfile[11] = "rFrame.pgm";
  char cFrfile[11] = "cFrame.pgm";

  int stfr=0; //start frame number
  int finfr=0; //end frame number
  int fr_nos[2] = {0,0}; //For frame numbers from command line

  FILE * fin;//input file
  FILE * fout;//output file

  FILE * rFout;//output file
  FILE * cFout;//output file

  int row = roww;
  int col = coll;
  int cb_row = cb_roww;
  int cb_col = cb_coll;
  int cr_row = cr_roww;
  int cr_col = cr_coll;

  int brow = broww;
  int bcol = bcoll;
  int rscope = rscopee;
  int cscope = cscopee;

  unsigned char lumaR[row][col];
  unsigned char lumaC[row][col];
  unsigned char cb[cb_row][cb_col];
  unsigned char cr[cr_row][cr_col];

  char xvec[row/brow][col/bcol];
  char yvec[row/brow][col/bcol];
  unsigned char lumaN[row][col];

  FILE * RecFileH;//output reconstructed frame
  char RecFrFile[13] = "RecFrame.pgm";

  int errFunc = 2; //1 for sad, 2 for mse
  //Variables Section Ends }

  strcpy(inpfile,def_in_file);
  strcpy(outfile,"out.dat");
  HandleArgs(argc,argv,inpfile,fr_nos);
  if(fr_nos[1] <= fr_nos[0]) PrintErr(fr_nos);
  //if(prn) printf("**************************************************\n");
  //if(prn) printf("INFO! Input File is = %s\n",argv[1]);
  //if(prn) printf("INFO! Reference Frame = %d\n",fr_nos[0]);
  //if(prn) printf("INFO! Current Frame = %d\n",fr_nos[1]);
  //if(prn) printf("**************************************************\n");
  if(prn) printf("**************************************************\n");
  if(prn) printf(" %s ",argv[1]);
  if(prn) printf(" Ref Frame = %d ",fr_nos[0]);
  if(prn) printf(" CurFrame = %d ",fr_nos[1]);
  if(prn) printf(" delta = %d ",fr_nos[1]-fr_nos[0]);
  //if(prn) printf("**************************************************\n");
  myfopen(&fin,0,inpfile); //0 means open as rb
  myfopen(&fout,2,outfile); //2 means open as w
  GetFrames(fin,lumaR,lumaC,cb,cr,row,col,cb_row,cb_col,cr_row,cr_col,fr_nos[0],fr_nos[1]);
  myfopen(&cFout,2,cFrfile); //2 means open as w
  PrintArray(cFout,lumaC,row,col);
  myfopen(&rFout,2,rFrfile); //2 means open as w
  PrintArray(rFout,lumaR,row,col);
  PrintInfo(fout,row,col,brow,bcol,rscope,cscope);
  GetVecs(lumaR,lumaC,fout,row,col,rscope,cscope,brow,bcol,xvec,yvec,errFunc);
  ReConstruct(xvec,
              yvec,
              lumaR,
              lumaN,
              row, 
              col, 
              brow, 
              bcol);
  myfopen(&RecFileH,2,RecFrFile); //2 means open as w
  PrintArray(RecFileH,lumaN,row,col);
  fclose(RecFileH);
  CalPsnr(lumaC,lumaN,row,col,bcol,brow,errFunc);

}// int main (int argc, char * argv[])


void HandleArgs(int argc, char * argv[], char * inpfile, int * fr_nos)
{
//###################################
//# This function Hs input args
//###################################
int stfr = 0;//local copy of it into this function.
int finfr = 0;//local copy of it into this function.
switch(argc)
  {
    case 1 :
      { 
        PrintHelpMess();
        exit(-1);
      }// case 1 :
    case 2 :
      { 
        //inpfile=argv[1];
        PrintHelpMess();
        exit(-1);
        break;
      }// case 2 :
    case 3 :
      {
        strcpy(inpfile,argv[1]);
        //printf("INFO! #1. Input file is %s\n",inpfile);
        //grab start frame and end frame numbers in variables.
        //The input argv is a string. so a string of type
        //say "23-45" which contains the start and end frame to be printed
        //must be processed to get integers stfr, start frame number
        //finfr, finish frame number.
        while(*argv[2] != '-')
        {
          stfr = *argv[2]-48+stfr; //Convert argv[2] to an integer stfr
          argv[2]++;
          stfr=stfr*10;
        }
        stfr=stfr/10; //Re-adjust the result
        argv[2]++;
        while(*argv[2])
        {
          finfr = *argv[2]-48+finfr;
          argv[2]++;
          finfr=finfr*10;
        }
        finfr=finfr/10; //Re-adjust the result
        fr_nos[0] = stfr;
        fr_nos[1] = finfr;
        break;
      }// case 3 :
    default: 
      {
        printf("ERROR! Too many args at command line. Exiting....\n");
        exit(-1);
      }
  }// switch(argc)
}// void HandleArgs(int argc, char * argv[], char * inpfile, int *fr_nos)
void PrintHelpMess()
{
        printf("###################################################\n");
        printf("#This Program extracts 2 named frames from qcif file,\n");
        printf("#Calculates motion estimation vectors,reconstructs \n");
        printf("#Frame using motion Vectors \n");
        printf("#And Finally calculates PSNR of the Reconstructed Frame\n");
        printf("###################################################\n");
        printf("Usage: unix> a.out <in file> <n1-n2>\n");
        printf("Where n1 is the reference frame number\n");
        printf("Where n2 is the frame to be encoded \n");
        printf("Example: unix> a.out bridge.qcif 45-47\n");
        printf("WARN!: Input args Format not checked. Pls conform to above shown usage\n");
        printf("###################################################\n");
}// void PrintHelpMess()
void myfopen(FILE **fileptr,int mode, char * fname)
{
//###################################
//# This function opens a file in
//# Binary Read Mode if mode = 0
//# Ascii Read Mode if mode = 1
//# Write Mode if mode = 2
//###################################
  //unsigned char luma[144][176];
  if(mode == 0)
  {
    if(!(*fileptr = fopen(fname, "rb")))
    {
      printf("1. Cannot open file :%s Exiting . . . .\n",fname);
      exit(-1);
    }// if(!(fileptr = fopen(fname, "r")))
  }// if(mode == 0)
  else if(mode == 1)
  {
    if(!(*fileptr = fopen(fname, "r")))
    {
      printf("2. Cannot open file :%s Exiting . . . .\n",fname);
      exit(-1);
    }// if(!(fileptr = fopen(fname, "r")))
  }// else if(mode == 1)
  else if(mode == 2)
  {
    if(!(*fileptr = fopen(fname, "w")))
    {
      printf("3. Cannot open file :%s Exiting . . . .\n",fname);
    exit(-1);
    }// if(!(fileptr = fopen(fname, "r")))
  }// else if(mode == 2)
  else
  {
    printf("############################################################\n");
    printf("ERROR! from function myfopen: Pls Dont Play with me\n");
    printf("Allowed Mode values are 0,1,2 where as mode passed is %d\n",mode);
    printf("Exiting....\n");
    printf("############################################################\n");
    exit(-1);
  }
}// void myfopen(FILE **fileptr,int mode, char * fname);

void GetFrames(
     FILE * fin,
     unsigned char lumaR[roww][coll],
     unsigned char lumaC[roww][coll],
     unsigned char cb[cb_roww][cb_coll],
     unsigned char cr[cr_roww][cr_coll],
     int row, 
     int col, 
     int cb_row,
     int cb_col,
     int cr_row,
     int cr_col,
     int RefFrNo,
     int CurrFrNo
     )
{
  int frno=0;
  int cc=0;
  int rr=0;
  for(frno=0;frno<RefFrNo+1;frno++)
  {
    fread(lumaR,1,row*col,fin); //read whole luma frame
    fread(cb,1,cb_row*cb_col,fin); //read whole luma frame
    fread(cr,1,cr_row*cb_col,fin); //read whole luma frame
  }// for(frno=0;frno<CurrFrno;frno++)
  for(;frno<CurrFrNo+1;frno++)
  {
    fread(lumaC,1,row*col,fin); //read whole luma frame
    fread(cb,1,cb_row*cb_col,fin); //read whole luma frame
    fread(cr,1,cr_row*cb_col,fin); //read whole luma frame
  }// for(frno=0;frno<CurrFrno;frno++)
}// void GetFrames()

void PrintArray(FILE * fr,
                unsigned char chararray[roww][coll],
                int row, 
                int col)
{
int rr=0;
int cc=0;
PrintHeader(fr,row,col);
for(rr=0;rr<row;rr++)
  for(cc=0;cc<col;cc++)
    fprintf(fr,"%d\n",chararray[rr][cc]);
}// void PrintArray()

void PrintHeader(FILE * fout,int row,int col)
{
  fprintf(fout,"P2\n");
  fprintf(fout,"%d %d\n",col,row);
  fprintf(fout,"255\n");
}// void PrintHeader(char * fname)

void PrintErr(int * fr_nos)
{
  printf("#################################################\n");
  printf("ERROR! Dont play around!!! Curr Frame Number %d must be greater than Ref Frame Number %d\n",fr_nos[1],fr_nos[0]);
  printf("Exiting....\n");
  printf("#################################################\n");
  exit(0);
}// void PrintErr(int * fr_nos)

void GetVecs(unsigned char lumaR[roww][coll],
             unsigned char lumaC[roww][coll],
             FILE * fout,
             int rows,
             int cols,
             int rscopes,
             int cscopes,
             int brows,
             int bcols,
             char xvec[roww/broww][coll/bcoll],
             char yvec[roww/broww][coll/bcoll],
             int errFunc)
{
  int rr=0;
  int cc=0;

  int iic=0;
  int iir=0;

  int row = rows;
  int col = cols;

  int brow = brows;
  int bcol = bcols;

  int refrow=0;
  int refcol=0;

  int mindex=0;

  int rscope = rscopes;
  int cscope = cscopes;

  int sad=0;
  int MinSad=0;
  int MinSadX=0;
  int MinSadY=0;

  int MSE = 0;
  int MinMSE = 0;
  int MinMSEX = 0;
  int MinMSEY = 0;
  
  int crow=0; //Centre of Diamond
  int ccol=0; //Centre of Diamond

  int refrr=0;
  int refcc=0;

  int it=0;
  int Minit=0;

  int MinXrr=0;
  int MinYrr=0;

  int rowfac=1;
  int colfac=1;

  int ii=0;
  int Minii=0;
  int frefrr;
  int frefcc;

  for(rr=0;rr<(row/brow);rr++)
  {
    for(cc=0;cc<(col/bcol);cc++)
    {
      crow=rr*brow;
      ccol=cc*bcol;
      rowfac=1;
      colfac=1;
      rscope = rscopee;
      cscope = cscopee;
      MinSad=1<<((sizeof(int)*8 -2)); //initialize MinSad to max +ive integer.
      MinMSE=1<<((sizeof(int)*8 -2)); //initialize MinSad to max +ive integer.
      ii=0;
        for(refrr=crow-brow;refrr<=crow+brow;refrr=refrr+brow)
        {
          for(refcc=ccol-bcol;refcc<=ccol+bcol;refcc=refcc+bcol)
          {
            mindex++;
            //ii++;
            if(refrr<0 || refrr>row-brow || refcc<0 || refcc>col-bcol){ii++; continue;}
            else ii++;
            frefrr=refrr;
            frefcc=refcc;
            //frefrr=modrefrr(refrr,ii,brow);
            //frefcc=modrefcc(refrr,ii,bcol);
            if(ii==1) frefrr = refrr+brow/2;
            if(ii==1) frefcc = refcc+bcol/2;
            if(ii==3) frefrr = refrr+brow/2;
            if(ii==3) frefcc = refcc-bcol/2;
            if(ii==7) frefrr = refrr-brow/2;
            if(ii==7) frefcc = refcc+bcol/2;
            if(ii==9) frefrr = refrr-brow/2;
            if(ii==9) frefcc = refcc-bcol/2;
            //debug printf("(%d,%d),refrr=%d,refcc=%d,frefrr=%d,frefcc=%d,crow=%d,ccol=%d,it=%d,ii=%d,stepr=%d,stepc=%d\n",rr,cc,refrr,refcc,frefrr,frefcc,crow,ccol,it,ii,brow/rowfac,bcol/colfac);
            MSE=0;
              for(iir=0;iir<brow;iir++)
              {
                for(iic=0;iic<bcol;iic++)
                {
                  unsigned char cVal=0;
                  unsigned char rVal=0;
                  unsigned char absdiff=0;
                  rVal = lumaR[frefrr+iir][frefcc+iic];
                  cVal = lumaC[(rr*brow)+iir][(cc*bcol)+iic];
                  absdiff = abs(rVal-cVal);
                  sad=sad+absdiff;
                  MSE=MSE+(cVal-rVal)*(cVal-rVal);
                }// for(iic=0;iic<bcol;iic++)
              }// for(iir=0;iir<brow;iir++)
              //debug printf("MSE=%d\n",MSE);
              if(MSE<MinMSE)
              {
                MinMSE = MSE;
                MinMSEX = frefrr-rr*brow;
                MinMSEY = frefcc-cc*bcol;
                MinXrr = refrr;
                MinYrr = refcc;
                Minii = ii;
                Minit = it;
              }// if(sad<MinSad)
          }// for(refcc=ccol-bcol/colfac;refcc<=ccol+bcol/colfac;refcc=refcc+bcol/colfac)
        }// for(refrr=crow-brow/rowfac;refrr<=crow+brow/rowfac;refrr=refrr+brow/rowfac)
        fprintf(fout,"(%d,%d) minMSE=%d,MSE Vectors: x=%d,y=%d,it=%d\n",rr,cc,MinMSE,MinMSEX,MinMSEY,it);
        //debug printf("(%d,%d) minMSE=%d,MSE Vectors: x=%d,y=%d,it=%d,Minii=%d,Minit=%d\n",rr,cc,MinMSE,MinMSEX,MinMSEY,it,Minii,Minit);
        xvec[rr][cc]=(unsigned char)MinMSEX; 
        yvec[rr][cc]=(unsigned char)MinMSEY;
        crow = rr*brow + MinMSEX;
        ccol = cc*bcol + MinMSEY;
        for(refrr=rr*brow-1;refrr<=rr*brow+1;refrr=refrr+1)
        {
          for(refcc=cc*bcol-1;refcc<=cc*bcol+1;refcc=refcc+1)
          {
              if(refrr<0 || refrr>row-brow || refcc<0 || refcc>col-bcol) continue;
              MSE=0;
              for(iir=0;iir<brow;iir++)
              {
                for(iic=0;iic<bcol;iic++)
                {
                  unsigned char cVal=0;
                  unsigned char rVal=0;
                  unsigned char absdiff=0;
                  rVal = lumaR[refrr+iir][refcc+iic];
                  cVal = lumaC[(rr*brow)+iir][(cc*bcol)+iic];
                  absdiff = abs(rVal-cVal);
                  sad=sad+absdiff;
                  MSE=MSE+(cVal-rVal)*(cVal-rVal);
                }// for(iic=0;iic<bcol;iic++)
              }// for(iir=0;iir<brow;iir++)
              if(MSE<MinMSE)
              {
                MinMSE = MSE;
                MinMSEX = refrr-rr*brow;
                MinMSEY = refcc-cc*bcol;
              }// if(sad<MinSad)
          }
        }
        fprintf(fout,"(%d,%d) minMSE=%d,MSE Vectors: x=%d,y=%d,it=%d\n",rr,cc,MinMSE,MinMSEX,MinMSEY,it);
        xvec[rr][cc]=(unsigned char)MinMSEX;
        yvec[rr][cc]=(unsigned char)MinMSEY;
        for(refrr=crow-1;refrr<=crow+1;refrr=refrr+1)
        {
          for(refcc=ccol-1;refcc<=ccol+1;refcc=refcc+1)
          {
              if(refrr<0 || refrr>row-brow || refcc<0 || refcc>col-bcol) continue;
              MSE=0;
              for(iir=0;iir<brow;iir++)
              {
                for(iic=0;iic<bcol;iic++)
                {
                  unsigned char cVal=0;
                  unsigned char rVal=0;
                  unsigned char absdiff=0;
                  rVal = lumaR[refrr+iir][refcc+iic];
                  //printf("indexrr = %d,index=%d, rVal=%d\n",refrr+iic,refcc+iic,rVal);
                  cVal = lumaC[(rr*brow)+iir][(cc*bcol)+iic];
                  absdiff = abs(rVal-cVal);
                  sad=sad+absdiff;
                  MSE=MSE+(cVal-rVal)*(cVal-rVal);
                }// for(iic=0;iic<bcol;iic++)
              }// for(iir=0;iir<brow;iir++)
              if(MSE<MinMSE)
              {
                MinMSE = MSE;
                MinMSEX = refrr-rr*brow;
                MinMSEY = refcc-cc*bcol;
              }// if(sad<MinSad)
          }// for(refcc=ccol-bcol/colfac;refcc<=ccol+bcol/colfac;refcc=refcc+bcol/colfac)
        }// for(refrr=crow-brow/rowfac;refrr<=crow+brow/rowfac;refrr=refrr+brow/rowfac)
        fprintf(fout,"(%d,%d) minMSE=%d,MSE Vectors: x=%d,y=%d,it=%d\n",rr,cc,MinMSE,MinMSEX,MinMSEY,it);
        xvec[rr][cc]=(unsigned char)MinMSEX;
        yvec[rr][cc]=(unsigned char)MinMSEY;
    }// for(cc=0;cc<(col/bcol);cc++)
  }// for(rr=0;rr<(row/brow);rr++)

  fclose(fout);
}// void GetVecs()

void PrintInfo(FILE *fout,
               int rows,
               int cols,
               int brows,
               int bcols,
               int rscopes,
               int cscopes)
{
  fprintf(fout,"The Frame has %d rows\n",rows);
  fprintf(fout,"The Frame has %d coloumns\n",cols);
  fprintf(fout,"Block is consist of %d rows,%d cols\n",brows,bcols); 
  fprintf(fout,"Current Frame has %d X %d = %d such blocks for which ME Vectors are Required\n",rows/brows,cols/bcols,rows*cols/(brows*bcols)); 
  fprintf(fout,"Hz search extends to +-%d pixels\n",cscopes); 
  fprintf(fout,"Vr search extends to +-%d pixels\n",rscopes); 
  fprintf(fout,"\n"); 
}// void PrintInfo(FILE *fout)

void ReConstruct(char xvec[roww/broww][coll/bcoll],
                char yvec[roww/broww][coll/bcoll],
                unsigned char lumaR[roww][coll],
                unsigned char lumaN[roww][coll], //Reconstructed Frame
                int rows, 
                int cols, 
                int brows, 
                int bcols
                )
{
  int rr=0;
  int cc=0;

  int iic=0;
  int iir=0;

  int row = rows;
  int col = cols;

  int brow = brows;
  int bcol = bcols;

  int refrow=0;
  int refcol=0;
  for(rr=0;rr<row/brow;rr++)
  {
    for(cc=0;cc<col/bcol;cc++)
    {
      for(iir=0;iir<brow;iir++)
      {
        for(iic=0;iic<bcol;iic++)
        {
          lumaN[rr*brow+iir][cc*bcol+iic]=
          lumaR[rr*brow+iir+xvec[rr][cc]][cc*bcol+iic+yvec[rr][cc]];
          //printf("xvec[%d][%d]=%d,yvec[%d][%d]=%d\n",rr,cc,xvec[rr][cc],rr,cc,yvec[rr][cc]);
        }// for(iic=0;iic<bcol;iic++)
      }// for(iir=0;iir<brow;iir++)
    }// for(cc=0;cc<col/bcol;cc++)
  }// for(rr=0;rr<row/brow;rr++)
}//void ReConsruct

void CalPsnr(
            unsigned char lumaC[roww][coll],
            unsigned char lumaN[roww][coll],
            int rows,
            int cols,
            int bcols,
            int brows,
            int errFunc
            )
{
  int rr=0;
  int cc=0;

  int row = rows;
  int col = cols;
  int brow = brows;
  int bcol = bcols;

  unsigned char refpel=0;
  unsigned char npel=0;

  int sqdiff = 0;

  double lpsnr=0.0;
  double psnr=0.0;

  double sigmadiff = 0.0;

  for(rr=0;rr<row;rr++)
  {
    for(cc=0;cc<col;cc++)
    {
      refpel = lumaC[rr][cc]; 
      npel = lumaN[rr][cc];
      sqdiff = (refpel-npel)*(refpel-npel);
      sigmadiff = sigmadiff + (double) sqdiff;
      //printf("sigmadiff = %f\n",sigmadiff);
    }// for(cc=0;cc<col/bcol;cc++)
  }// for(rr=0;rr<row/brow;rr++)
  sigmadiff = sigmadiff/(double)(row*col);
  //printf("sigmadiff final = %f\n",sigmadiff);
  psnr = ((255*255)/sigmadiff);
  lpsnr = 10* (log(psnr)/log(10));
  if(errFunc==1) printf("sad ");
  else printf("mse ");
  printf("psnr=%f\n",lpsnr);
}// void CalPsnr();
