////////////////////////////////////////////////////////////////////////////
////
//Author Aviral Mittal
//ISLI 2005/2006
////////////////////////////////////////////////////////////////////////////
////
`timescale 1ns/1ps
//`include "./functions.v"

module mse(mv_x,
           mv_y,
           min_sad,
           partition_mode,
           mfinish,
           ADRRef,
           CENRef,
           WENRef,
           OENRef,
           ADRCur,
           CENCur,
           WENCur,
           OENCur,
           ADRNew ,
           CENNew ,
           WENNew,
           OENNew,
           Dnew,
           start,
           clk,
           reset,
           Dref,
           Dcur
           );

`include "./functions.v"

//parameters Declaration Section
parameter roww = 144; //Qcif/Cif rows
parameter l2roww = 9; //Qcif/Cif rows register width max can be 288
parameter coll = 176; //Qcif/Cif columns
parameter l2coll = 9; //Qcif/Cif columns register width max can be 352

parameter broww = 4;  //Initial Block Size row
parameter l2broww = 5;  //Block Size row reg max value can be 32
parameter bcoll = 4;  //Initial Block Size column
parameter l2bcoll = 5;  //Block Size column reg max value can be 32

parameter rscopee = 8; //Initial search size row
parameter l2rscopee = 6; //Initial search size row max value = 64
parameter cscopee = 8; //Initial search size column
parameter l2cscopee = 6; //Initial search size column max value = 64

parameter rrwidth = 7;// max value of rr is max(roww)/min(broww) 288/4=72
parameter ccwidth = 7;// max value of rr is max(coll)/min(bcoll) 352/4=88

parameter cif_qcif = 4; //Make it 5 for cif, 4 for qcif


//parameters Declaration Section Ends
  input start;
  input clk;
  input reset;
  input [7:0] Dref;
  input [7:0] Dcur;

  output [7:0]  mv_x;
  output [7:0]  mv_y;
  output [15:0] min_sad;
  output [9:0]  partition_mode;
  output        mfinish;

  output [16:0] ADRRef;
  output CENRef;
  output WENRef;
  output OENRef;

  output [16:0] ADRCur;
  output CENCur;
  output WENCur;
  output OENCur;

  output [16:0] ADRNew;
  output CENNew;
  output WENNew;
  output OENNew;

  output [7:0] Dnew;


  reg [rrwidth-1:0] rr;
  reg [rrwidth-1:0] n_rr;
  reg [ccwidth-1:0] cc;
  reg [ccwidth-1:0] n_cc;

  reg [l2broww-1:0] brow;
  reg [l2broww-1:0] n_brow;
  reg [l2bcoll-1:0] bcol;
  reg [l2bcoll-1:0] n_bcol;

  reg [l2rscopee-1:0] rscope;
  reg [l2rscopee-1:0] n_rscope;
  reg [l2cscopee-1:0] cscope;
  reg [l2cscopee-1:0] n_cscope;

  reg [l2broww-1:0] iir;
  reg [l2broww-1:0] n_iir;
  reg [l2bcoll-1:0] iic;
  reg [l2bcoll-1:0] n_iic;

  reg [l2roww-1:0] refrr;
  reg [l2roww-1:0] n_refrr;
  reg [l2coll-1:0] refcc;
  reg [l2coll-1:0] n_refcc;
  reg [l2coll-1:0] frefrr;
  reg [l2coll-1:0] frefcc;

  reg [1:0] iter;
  reg [1:0] n_iter;

  reg [3:0] ii;
  reg [3:0] n_ii;

  reg [17:0] MSE;
  reg [17:0] MinMSE;
  reg [l2rscopee-1:0] MinMSEx;
  reg [l2cscopee-1:0] MinMSEy;
  reg [l2rscopee-1:0] xvec;
  reg [l2cscopee-1:0] yvec;

  wire [16:0] refAddr;
  wire [16:0] curAddr;
 
  wire [l2roww-1:0] crow;
  wire [l2coll-1:0] ccol;

  reg [rrwidth-1:0] n_crow;
  reg [rrwidth-1:0] n_ccol;
 
  reg ld_crow;
  reg ld_ccol;

  wire row_err;
  wire col_err;

  wire addr_err;

  reg max_iic_del; //For generating a puslse of iic==brow, to make sure
                   //this sinal is '1' for only one clock cycle
                   //This will 'stall' the design properly
                   //without if, if start goes to '0', when iic==3,
                   //The design will keep advancing
  wire max_iic_pulse; //For generating a puslse of iic==brow, to make sure
                   //this sinal is '1' for only one clock cycle
                   //This will 'stall' the design properly
                   //without if, if start goes to '0', when iic==3,
                   //The design will keep advancing
  reg start_del;
  wire start_pulse;

  reg mindex;
  reg n_mindex;
  wire OneMatchDone;
  wire OneSearchComplete;
  wire OneIterDone;
  wire OneFrameDone;

  wire incr_iir;
  wire incr_iic;
  wire incr_rr;
  wire incr_cc;
  wire incr_refrr;
  wire incr_refcc;
  wire incr_iter;
  wire incr_ii;
  wire incr_mindex;
  wire ld_refrr;
  wire ld_refcc;

  wire max_iir;
  wire max_iic;
  wire max_rr;
  wire max_cc;
  wire max_refrr;
  wire max_refcc;
  wire max_iter;
  wire max_ii;

//   function [16:0] mlog;
//   input [4:0] mvec;
//   begin
//     case(mvec)
//       4 : mlog = 2;
//       8 : mlog = 3;
//       16 : mlog = 4;
//       32 : mlog = 5;
//       default : mlog=0;
//     endcase
//   end
//   endfunction
  

//Reset all registers  
  always @ (posedge clk)
  begin
    if(reset == 0)
    begin
      rr <= 0;
      cc <= 0;
      iir <= 0;
      iic <= 0;
      brow <= broww;
      bcol <= bcoll;
      refrr <= 0;
      refcc <= 0;
      ii <= 0;
      iter <= 0;
      //crow <= 0;
      //ccol <= 0;
      max_iic_del <= 0;
      start_del <= 0;
    end
    else
    begin
      rr <= n_rr;
      cc <= n_cc;
      iir <= n_iir;
      iic <= n_iic;
      refrr <= n_refrr;
      refcc <= n_refcc;
      ii <= n_ii;
      iter <= n_iter;
      //crow <= n_crow;
      //ccol <= n_ccol;
      max_iic_del <= max_iic;
      start_del <= start;
    end 
  end
//Reset all registers Ends

  assign start_pulse = start & ~start_del;

// Combo Logic for rr

  always @ (iic or incr_iic)
  begin
    n_iic <= iic;
    if(incr_iic)
      if(iic == bcoll -1)
        n_iic <= 0;
      else
        n_iic <= iic + 1;
  end

  always @ (iir or incr_iir)
  begin
    n_iir <= iir;
      if(incr_iir)
        if(iir == broww -1)
          n_iir <= 0;
        else
          n_iir <= iir + 1;
  end

  always @ (rr or incr_rr)
  begin
    n_rr <= rr;
    if(incr_rr)
      if(rr == roww/broww -1)
        n_rr <= 0;
      else
        n_rr <= rr + 1;
  end


  always @ (cc or incr_cc)
  begin
    n_cc <= cc;
    if(incr_cc)
      if(cc == coll/bcoll -1)
        n_cc <= 0;
      else
        n_cc <= cc + 1;
  end

  always @ (ii or incr_ii)
  begin
    n_ii <= ii;
    if(incr_ii)
      if(ii != 8)
        n_ii <= ii + 1;
      else
        n_ii <= 0;
  end


// if(ii==1) frefrr = refrr+brow/2;
// if(ii==1) frefcc = refcc+bcol/2;
// if(ii==3) frefrr = refrr+brow/2;
// if(ii==3) frefcc = refcc-bcol/2;
// if(ii==7) frefrr = refrr-brow/2;
// if(ii==7) frefcc = refcc+bcol/2;
// if(ii==9) frefrr = refrr-brow/2;
// if(ii==9) frefcc = refcc-bcol/2;

  //refrr,refcc
  assign crow = rr<<(mlog(brow));
  assign ccol = cc<<(mlog(bcol));
  always @ (refrr or incr_refrr or n_iter or n_ii or iir or crow or brow or ld_refrr)
  begin
    n_refrr <= refrr;
    if(ld_refrr == 1 && n_iter == 0 && n_ii==0)
      n_refrr <= crow-brow;
    else if(ld_refrr == 1 && n_iter != 0 && n_ii==0)
      n_refrr <= crow-1;
    else if(incr_refrr == 1 && n_iter == 0)
      if(refrr == crow+brow)
        n_refrr <= crow-brow;
      else
        n_refrr <= refrr+brow;
    else if(incr_refrr && n_iter !=0) 
      if(refrr == crow+1)
        n_refrr <= crow-1;
      else
        n_refrr <= refrr+1;
  end

  always @ (refrr or ii or iter or brow)
  begin
    frefrr <= refrr;
    if((ii==0 || ii==2) && iter==0)
      frefrr <= refrr+(brow>>1);
    else if((ii==6 || ii==8) && iter==0)
      frefrr <= refrr-(brow>>1);
  end

  always @ (refcc or incr_refcc or n_iter or n_ii or iir or ccol or bcol or ld_refcc)
  begin
    n_refcc <= refcc;
    if(ld_refcc == 1 && n_iter == 0 && n_ii==0)
      n_refcc <= ccol-bcol;
    else if(ld_refcc == 1 && n_iter != 0 && n_ii==0)
      n_refcc <= ccol-1;
    else if(incr_refcc == 1 && n_iter == 0)
      if(refcc == ccol+bcol)
        n_refcc <= ccol-bcol;
      else
        n_refcc <= refcc+bcol;
    else if(incr_refcc && n_iter != 0) 
      if(refcc == ccol+1)
        n_refcc <= ccol-1;
      else
        n_refcc <= refcc+1;
  end

  always @ (refcc or ii or iter or bcol)
  begin
    frefcc <= refcc;
    if((ii==0 || ii==6) && (iter==0))
      frefcc <= refcc+(bcol>>1);
    else if((ii==2 || ii==8) && (iter==0))
      frefcc <= refcc-(bcol>>1);
  end


//  always @(ccol or ld_ccol)
//  begin
//    n_ccol <= ccol;
//    if(ld_ccol)
//      n_ccol <= cc*bcol;
//  end

//  always @(crow or ld_crow)
//  begin
//    n_crow <= crow;
//    if(ld_crow)
//      n_crow <= rr*brow;
//  end

  always @(iter or incr_iter)
  begin
    n_iter <= iter;
    if(incr_iter)
      if(iter == 2)
        n_iter <= 0;
      else
        n_iter <= iter+1;
  end

  always @ (OneMatchDone or mindex)
  begin
    n_mindex <= mindex;
    if(OneMatchDone && mindex == 26)
      n_mindex <= 0;
    else if(OneMatchDone)
      n_mindex <= n_mindex+1;
  end


  assign max_iic = (iic==bcol-1) ? 1 : 0;
  assign max_iir = (iir==brow-1) ? 1 : 0;
  assign max_iter = (iter==3-1) ? 1 : 0;
  assign max_ii = (ii==9-1) ? 1 : 0;

  assign max_cc = (cc==(coll>>mlog(bcol))-1) ? 1 : 0;
  assign max_rr = (rr==(roww>>mlog(brow))-1) ? 1 : 0;

  assign max_iic_pulse = max_iic & ~max_iic_del;
  assign incr_iic=start_del;
  assign incr_iir = max_iic_pulse;

  assign incr_ii   = OneMatchDone;
  assign incr_iter = OneIterDone;

//  always @(max_iter or max_ii or max_refcc or max_refrr or max_refrr or max_iir or max_iic)
//  beign
//    incr_iir = 0;
//    incr_refrr = 0;
//    incr_refcc = 0;
//    incr_iter = 0;
//    incr_ii = 0;
//    incr_rr = 0;
//    incr_cc = 0;
//    if(incr_iic)
//    begin
//    end
//        
//  end

  assign OneMatchDone = max_iic_pulse & max_iir;
  assign OneIterDone  = OneMatchDone & max_ii;
  assign OneSearchComplete = OneIterDone & max_iter;
  assign incr_cc = OneSearchComplete;
  assign incr_rr = incr_cc & max_cc;
  assign OneFrameDone = incr_rr & max_rr;


  assign ld_refcc = start_pulse | OneMatchDone; 
  assign ld_refrr = start_pulse | OneMatchDone; 

  assign incr_refcc = OneMatchDone;
  assign incr_refrr = OneMatchDone & (ii==2 || ii==5 || ii==8);

  assign row_err = (refrr[l2roww-1]) ? 1 : 0;
  assign col_err = (refcc[l2coll-1]) ? 1 : 0;


  //ACU
  //addr = ((rr*brow)+iir)*col+(cc*bcol)+iic); //from Reference C Model
  assign refAddr = ((((rr<<mlog(brow))|iir)+ (((rr<<mlog(brow))|iir)<<1)+(((rr<<mlog(brow))|iir)<<3))<<cif_qcif) + ((cc<<mlog(bcol))|iic);

  assign curAddr = (((frefrr+iir)+ ((frefrr+iir)<<1)+((frefrr+iir)<<3))<<cif_qcif) + (frefcc+iic);

  assign addr_err = row_err | col_err;

  assign ADRRef  = refAddr;
  assign ADRCur = (addr_err) ? 0 : curAddr;
 
  assign CENRef = ~start | start_pulse;
  assign OENRef = ~start | start_pulse;
  assign WENRef = 1'b1;

  assign CENCur = ~start | start_pulse;
  assign OENCur = ~start | start_pulse;
  assign WENCur = 1'b1;
endmodule
