////////////////////////////////////////////////////////////////////////////
////
//Author Aviral Mittal
//ISLI 2005/2006
////////////////////////////////////////////////////////////////////////////
////
`timescale 1ns/1ps
//`include "./functions.v"

module mse(start,clk,reset);
//parameters Declaration Section
parameter roww = 144; //Qcif/Cif rows
parameter l2roww = 9; //Qcif/Cif rows register width max can be 288
parameter coll = 176; //Qcif/Cif columns
parameter l2coll = 9; //Qcif/Cif columns register width max can be 352

parameter broww = 4;  //Initial Block Size row
parameter l2broww = 5;  //Block Size row reg max value can be 32
parameter bcoll = 4;  //Initial Block Size column
parameter l2bcoll = 5;  //Block Size column reg max value can be 32

parameter rscopee = 8; //Initial search size row
parameter l2rscopee = 6; //Initial search size row max value = 64
parameter cscopee = 8; //Initial search size column
parameter l2cscopee = 6; //Initial search size column max value = 64

parameter rrwidth = 7;// max value of rr is max(roww)/min(broww) 288/4=72
parameter ccwidth = 7;// max value of rr is max(coll)/min(bcoll) 352/4=88

parameter cif_qcif = 4; //Make it 5 for cif, 4 for qcif


//parameters Declaration Section Ends
  input start;
  input clk;
  input reset;


  reg [rrwidth-1:0] rr;
  reg [rrwidth-1:0] n_rr;
  reg [ccwidth-1:0] cc;
  reg [ccwidth-1:0] n_cc;

  reg [l2broww-1:0] brow;
  reg [l2broww-1:0] n_brow;
  reg [l2bcoll-1:0] bcol;
  reg [l2bcoll-1:0] n_bcol;

  reg [l2rscopee-1:0] rscope;
  reg [l2rscopee-1:0] n_rscope;
  reg [l2cscopee-1:0] cscope;
  reg [l2cscopee-1:0] n_cscope;

  reg [l2broww-1:0] iir;
  reg [l2broww-1:0] n_iir;
  reg [l2bcoll-1:0] iic;
  reg [l2bcoll-1:0] n_iic;

  reg [rrwidth-1:0] refrr;
  reg [rrwidth-1:0] n_refrr;
  reg [ccwidth-1:0] refcc;
  reg [rrwidth-1:0] n_refcc;
  reg ld_refrr;
  reg ld_refcc;

  reg [1:0] iter;
  reg [1:0] n_iter;
  reg incr_iter;

  reg [3:0] ii;
  reg [3:0] n_ii;
  reg incr_ii;

  reg [17:0] MSE;
  reg [17:0] MinMSE;
  reg [l2rscopee-1:0] MinMSEx;
  reg [l2cscopee-1:0] MinMSEy;
  reg [l2rscopee-1:0] xvec;
  reg [l2cscopee-1:0] yvec;

  wire [16:0] refAddr;
 
  reg [roww/broww:0] crow;
  reg [coll/bcoll:0] ccol;

  reg [roww/broww:0] n_crow;
  reg [coll/bcoll:0] n_ccol;
 
  reg ld_crow;
  reg ld_ccol;

  reg row_err;
  reg col_err;

  reg mindex;
  reg n_mindex;
  wire OneMatchDone;

   function [16:0] mlog;
   input [4:0] mvec;
   begin
     case(mvec)
       4 : mlog = 2;
       8 : mlog = 3;
       16 : mlog = 4;
       32 : mlog = 5;
       default : mlog=0;
     endcase
   end
   endfunction
  

//Reset all registers  
  always @ (posedge clk)
  begin
    if(reset == 0)
    begin
      rr <= 0;
      cc <= 0;
      iir <= 0;
      iic <= 0;
      brow <= broww;
      bcol <= bcoll;
      refrr <= 0;
      refcc <= 0;
      ii <= 0;
      iter <= 0;
      crow <= 0;
      ccol <= 0;
    end
    else
    begin
      rr <= n_rr;
      cc <= n_cc;
      iir <= n_iir;
      iic <= n_iic;
      refrr <= n_refrr;
      refcc <= n_refcc;
      ii <= n_ii;
      iter <= n_iter;
      crow <= n_crow;
      ccol <= n_ccol;
    end 
  end
//Reset all registers Ends

// Combo Logic for rr

  always @ (iic)
  begin
    n_iic <= iic;
    if(iic == bcoll -1)
      n_iic <= 0;
    else
      n_iic <= iic + 1;
  end

  always @ (iir,iic)
  begin
    n_iir <= iir;
    if(iic==bcoll-1)
      if(iir == broww -1)
        n_iir <= 0;
      else
        n_iir <= iir + 1;
  end

  always @ (cc,iir,iic,OneMatchDone)
  begin
    n_cc <= cc;
    if(iic==bcoll-1 && iir==broww-1 && OneMatchDone)
      if(cc == coll/bcoll -1)
        n_cc <= 0;
      else
        n_cc <= cc + 1;
  end

  always @ (rr,cc,iir,iic,OneMatchDone)
  begin
    n_rr <= rr;
    //if(iic==bcoll-1 && iir==broww-1 && cc==coll/bcoll-1)
    if(iir==broww-1 && cc==coll/bcoll-1 && OneMatchDone)
      if(rr == roww/broww -1)
        n_rr <= 0;
      else if(OneMatchDone)
        n_rr <= rr + 1;
  end


  always @ (ii,incr_ii)
  begin
    n_ii <= ii;
    if(incr_ii && ii != 8)
      n_ii <= ii + 1;
    else if(incr_ii && ii == 8)
      n_ii <= 0;
  end

  //refrr,refcc
  //assign crow = rr*brow;
  //assign ccol = cc*bcol;
  always @ (refrr,ld_refrr,iter,ii,iir)
  begin
    n_refrr <= refrr;
    row_err <= 0;
    if(ld_refrr == 1 && iter == 0 && ii==0)
    begin
      n_refrr <= crow-brow;
      if(crow-brow<0) row_err <= 1;
    end
    else if(ld_refrr == 1 && iter != 0 && ii==0)
    begin
      n_refrr <= crow-1;
      if(crow-1<0) row_err <= 1;
    end
    else if(ld_refrr == 1 && iter == 0 && iir == brow-1)
      n_refrr <= n_refrr+brow;
    else if(ld_refrr && iir==brow-1) 
      n_refrr <= refrr+1;
  end

  always @ (refcc,ld_refcc,iter,ii,iir)
  begin
    n_refcc <= refcc;
    col_err <= 0;
    if(ld_refcc == 1 && iter == 0 && ii==0)
    begin
      n_refcc <= ccol-bcol;
      if(ccol-bcol<0) col_err <= 1;
    end
    else if(ld_refcc == 1 && iter != 0 && ii==0)
    begin
      n_refcc <= ccol-1;
      if(ccol-1<0) col_err <= 1;
    end
    else if(ld_refcc == 1 && iter == 0 && iir == brow-1)
      n_refcc <= n_refcc+bcol;
    else if(ld_refcc && iir == brow-1) 
      n_refcc <= n_refcc+1;
  end

  always @(ccol,ld_ccol)
  begin
    n_ccol <= ccol;
    if(ld_ccol)
      n_ccol <= cc*bcol;
  end

  always @(crow,ld_crow)
  begin
    n_crow <= crow;
    if(ld_crow)
      n_crow <= rr*brow;
  end

  always @(iter,incr_iter)
  begin
    n_iter <= iter;
    if(iter == 2 && incr_iter)
      n_iter <= 0;
    else if(incr_iter)
      n_iter <= iter+1;
  end

  assign OneMatchDone = (iir==brow-1)? 1'b1:1'b0;
  always @ (OneMatchDone,mindex)
  begin
    n_mindex <= mindex;
    if(OneMatchdone && m_index == 26)
      n_mindex <= 0;
    else if(OneMatchDone)
      n_mindex <= n_mindex+1;
  end


  //ACU
  //addr = ((rr*brow)+iir)*col+(cc*bcol)+iic); //from Reference C Model
  assign refAddr = ((((rr<<mlog(brow))|iir)+ (((rr<<mlog(brow))|iir)<<1)+(((rr<<mlog(brow))|iir)<<3))<<cif_qcif) + ((cc<<mlog(bcol))|iic);

endmodule
